@@ -158,12 +158,11 @@ verify_as_reply(krb5_context            context,
      * principal) and we requested (and received) a TGT.
      */
     canon_req = ((request->kdc_options & KDC_OPT_CANONICALIZE) != 0) ||
-        (krb5_princ_type(context, request->client) ==
-         KRB5_NT_ENTERPRISE_PRINCIPAL) ||
+        request->client->type == KRB5_NT_ENTERPRISE_PRINCIPAL ||
         (request->kdc_options & KDC_OPT_REQUEST_ANONYMOUS);
     if (canon_req) {
-        canon_ok = IS_TGS_PRINC(context, request->server) &&
-            IS_TGS_PRINC(context, as_reply->enc_part2->server);
+        canon_ok = IS_TGS_PRINC(request->server) &&
+            IS_TGS_PRINC(as_reply->enc_part2->server);
         if (!canon_ok && (request->kdc_options & KDC_OPT_REQUEST_ANONYMOUS)) {
             canon_ok = krb5_principal_compare_any_realm(context,
                                                         as_reply->client,
@@ -212,8 +211,6 @@ verify_as_reply(krb5_context            context,
 
 static krb5_error_code
 stash_as_reply(krb5_context             context,
-               krb5_timestamp           time_now,
-               krb5_kdc_req             *request,
                krb5_kdc_rep             *as_reply,
                krb5_creds *             creds,
                krb5_ccache              ccache)
@@ -475,11 +472,8 @@ build_in_tkt_name(krb5_context context,
      * Windows Server 2008 R2 RODC insists on TGS principal names having the
      * right name type.
      */
-    if (krb5_princ_size(context, server) == 2 &&
-        data_eq_string(*krb5_princ_component(context, server, 0),
-                       KRB5_TGS_NAME)) {
-        krb5_princ_type(context, server) = KRB5_NT_SRV_INST;
-    }
+    if (server->length == 2 && data_eq_string(server->data[0], KRB5_TGS_NAME))
+        server->type = KRB5_NT_SRV_INST;
     *server_out = server;
     return 0;
 }
@@ -491,14 +485,14 @@ krb5_init_creds_free(krb5_context context,
     if (ctx == NULL)
         return;
 
-    if (ctx->opte != NULL && krb5_gic_opt_is_shadowed(ctx->opte)) {
+    if (ctx->opte != NULL && gic_opt_is_shadowed(ctx->opte)) {
         krb5_get_init_creds_opt_free(context,
                                      (krb5_get_init_creds_opt *)ctx->opte);
     }
     k5_response_items_free(ctx->rctx.items);
     free(ctx->in_tkt_service);
-    zap(ctx->password.data, ctx->password.length);
-    krb5_free_data_contents(context, &ctx->password);
+    zapfree(ctx->gakpw.storage.data, ctx->gakpw.storage.length);
+    k5_preauth_request_context_fini(context);
     krb5_free_error(context, ctx->err_reply);
     krb5_free_pa_data(context, ctx->err_padata);
     krb5_free_cred_contents(context, &ctx->cred);
@@ -586,7 +580,7 @@ krb5_init_creds_get_creds(krb5_context context,
     if (!ctx->complete)
         return KRB5_NO_TKT_SUPPLIED;
 
-    return krb5int_copy_creds_contents(context, &ctx->cred, creds);
+    return k5_copy_creds_contents(context, &ctx->cred, creds);
 }
 
 krb5_error_code KRB5_CALLCONV
@@ -656,6 +650,74 @@ cleanup:
     return code;
 }
 
+/* Return the current time, possibly using the offset from a previously
+ * received preauth-required error. */
+krb5_error_code
+k5_init_creds_current_time(krb5_context context, krb5_init_creds_context ctx,
+                           krb5_boolean allow_unauth, krb5_timestamp *time_out,
+                           krb5_int32 *usec_out)
+{
+    if (ctx->pa_offset_state != NO_OFFSET &&
+        (allow_unauth || ctx->pa_offset_state == AUTH_OFFSET) &&
+        (context->library_options & KRB5_LIBOPT_SYNC_KDCTIME)) {
+        /* Use the offset we got from a preauth-required error. */
+        return k5_time_with_offset(ctx->pa_offset, ctx->pa_offset_usec,
+                                   time_out, usec_out);
+    } else {
+        /* Use the time offset from the context, or no offset. */
+        return krb5_us_timeofday(context, time_out, usec_out);
+    }
+}
+
+/* Choose a random nonce for ctx->request. */
+static krb5_error_code
+pick_nonce(krb5_context context, krb5_init_creds_context ctx)
+{
+    krb5_error_code code = 0;
+    unsigned char random_buf[4];
+    krb5_data random_data = make_data(random_buf, 4);
+
+    /* We incorrectly encode this as signed, so make sure we use an unsigned
+     * value to avoid interoperability issues. */
+    code = krb5_c_random_make_octets(context, &random_data);
+    if (code != 0)
+        return code;
+    ctx->request->nonce = 0x7fffffff & load_32_n(random_buf);
+    return 0;
+}
+
+/* Set the timestamps for ctx->request based on the desired lifetimes. */
+static krb5_error_code
+set_request_times(krb5_context context, krb5_init_creds_context ctx)
+{
+    krb5_error_code code;
+    krb5_timestamp from, now;
+    krb5_int32 now_ms;
+
+    code = k5_init_creds_current_time(context, ctx, TRUE, &now, &now_ms);
+    if (code != 0)
+        return code;
+
+    /* Omit request start time unless the caller explicitly asked for one. */
+    from = krb5int_addint32(now, ctx->start_time);
+    if (ctx->start_time != 0)
+        ctx->request->from = from;
+
+    ctx->request->till = krb5int_addint32(from, ctx->tkt_life);
+
+    if (ctx->renew_life > 0) {
+        /* Don't ask for a smaller renewable time than the lifetime. */
+        ctx->request->rtime = krb5int_addint32(from, ctx->renew_life);
+        if (ctx->request->rtime < ctx->request->till)
+            ctx->request->rtime = ctx->request->till;
+        ctx->request->kdc_options &= ~KDC_OPT_RENEWABLE_OK;
+    } else {
+        ctx->request->rtime = 0;
+    }
+
+    return 0;
+}
+
 /**
  * Throw away any state related to specific realm either at the beginning of a
  * request, or when a realm changes, or when we start to use FAST after
@@ -670,9 +732,6 @@ restart_init_creds_loop(krb5_context context, krb5_init_creds_context ctx,
                         krb5_pa_data **padata)
 {
     krb5_error_code code = 0;
-    unsigned char random_buf[4];
-    krb5_data random_data;
-    krb5_timestamp from;
 
     if (ctx->preauth_to_use) {
         krb5_free_pa_data(context, ctx->preauth_to_use);
@@ -686,8 +745,7 @@ restart_init_creds_loop(krb5_context context, krb5_init_creds_context ctx,
     code = krb5int_fast_make_state(context, &ctx->fast_state);
     if (code != 0)
         goto cleanup;
-    ctx->preauth_rock.fast_state = ctx->fast_state;
-    krb5_preauth_request_context_init(context);
+    k5_preauth_request_context_init(context);
     if (ctx->outer_request_body) {
         krb5_free_data(context, ctx->outer_request_body);
         ctx->outer_request_body = NULL;
@@ -700,18 +758,6 @@ restart_init_creds_loop(krb5_context context, krb5_init_creds_context ctx,
             goto cleanup;
     }
 
-    /* Set the request nonce. */
-    random_data.length = 4;
-    random_data.data = (char *)random_buf;
-    code = krb5_c_random_make_octets(context, &random_data);
-    if (code !=0)
-        goto cleanup;
-    /*
-     * See RT ticket 3196 at MIT.  If we set the high bit, we may have
-     * compatibility problems with Heimdal, because we (incorrectly) encode
-     * this value as signed.
-     */
-    ctx->request->nonce = 0x7fffffff & load_32_n(random_buf);
     krb5_free_principal(context, ctx->request->server);
     ctx->request->server = NULL;
 
@@ -721,8 +767,6 @@ restart_init_creds_loop(krb5_context context, krb5_init_creds_context ctx,
     if (code != 0)
         goto cleanup;
 
-    ctx->request_time = time(NULL);
-
     code = krb5int_fast_as_armor(context, ctx->fast_state,
                                  ctx->opte, ctx->request);
     if (code != 0)
@@ -734,25 +778,8 @@ restart_init_creds_loop(krb5_context context, krb5_init_creds_context ctx,
             goto cleanup;
     }
     /* give the preauth plugins a chance to prep the request body */
-    krb5_preauth_prepare_request(context, ctx->opte, ctx->request);
+    k5_preauth_prepare_request(context, ctx->opte, ctx->request);
 
-    /* Omit request start time in the common case.  MIT and Heimdal KDCs will
-     * ignore it for non-postdated tickets anyway. */
-    from = krb5int_addint32(ctx->request_time, ctx->start_time);
-    if (ctx->start_time != 0)
-        ctx->request->from = from;
-    ctx->request->till = krb5int_addint32(from, ctx->tkt_life);
-
-    if (ctx->renew_life > 0) {
-        ctx->request->rtime =
-            krb5int_addint32(from, ctx->renew_life);
-        if (ctx->request->rtime < ctx->request->till) {
-            /* don't ask for a smaller renewable time than the lifetime */
-            ctx->request->rtime = ctx->request->till;
-        }
-        ctx->request->kdc_options &= ~(KDC_OPT_RENEWABLE_OK);
-    } else
-        ctx->request->rtime = 0;
     code = krb5int_fast_prep_req_body(context, ctx->fast_state,
                                       ctx->request,
                                       &ctx->outer_request_body);
@@ -795,9 +822,8 @@ krb5_init_creds_init(krb5_context context,
     ctx->prompter = prompter;
     ctx->prompter_data = data;
     ctx->gak_fct = krb5_get_as_key_password;
-    ctx->gak_data = &ctx->password;
+    ctx->gak_data = &ctx->gakpw;
 
-    ctx->request_time = 0; /* filled in later */
     ctx->start_time = start_time;
 
     if (options == NULL) {
@@ -811,8 +837,8 @@ krb5_init_creds_init(krb5_context context,
         options = &local_opts;
     }
 
-    code = krb5int_gic_opt_to_opte(context, options,
-                                   &ctx->opte, 1, "krb5_init_creds_init");
+    code = k5_gic_opt_to_opte(context, options, &ctx->opte, 1,
+                              "krb5_init_creds_init");
     if (code != 0)
         goto cleanup;
 
@@ -822,23 +848,6 @@ krb5_init_creds_init(krb5_context context,
 
     opte = ctx->opte;
 
-    ctx->preauth_rock.magic = CLIENT_ROCK_MAGIC;
-    ctx->preauth_rock.etype = &ctx->etype;
-    ctx->preauth_rock.as_key = &ctx->as_key;
-    ctx->preauth_rock.gak_fct = &ctx->gak_fct;
-    ctx->preauth_rock.gak_data = &ctx->gak_data;
-    ctx->preauth_rock.default_salt = &ctx->default_salt;
-    ctx->preauth_rock.salt = &ctx->salt;
-    ctx->preauth_rock.s2kparams = &ctx->s2kparams;
-    ctx->preauth_rock.rctx = ctx->rctx;
-    ctx->preauth_rock.client = client;
-    ctx->preauth_rock.prompter = prompter;
-    ctx->preauth_rock.prompter_data = data;
-    ctx->preauth_rock.allowed_preauth_type = &ctx->allowed_preauth_type;
-    ctx->preauth_rock.selected_preauth_type = &ctx->selected_preauth_type;
-    ctx->preauth_rock.cc_config_in = &ctx->cc_config_in;
-    ctx->preauth_rock.cc_config_out = &ctx->cc_config_out;
-
     /* Initialise request parameters as per krb5_get_init_creds() */
     ctx->request->kdc_options = context->kdc_default_options;
 
@@ -911,13 +920,11 @@ krb5_init_creds_init(krb5_context context,
     /* enctypes */
     if (opte->flags & KRB5_GET_INIT_CREDS_OPT_ETYPE_LIST) {
         ctx->request->ktype =
-            k5alloc((opte->etype_list_length * sizeof(krb5_enctype)),
-                    &code);
+            k5memdup(opte->etype_list,
+                     opte->etype_list_length * sizeof(krb5_enctype), &code);
         if (code != 0)
             goto cleanup;
         ctx->request->nktypes = opte->etype_list_length;
-        memcpy(ctx->request->ktype, opte->etype_list,
-               ctx->request->nktypes * sizeof(krb5_enctype));
     } else if (krb5_get_default_in_tkt_ktypes(context,
                                               &ctx->request->ktype) == 0) {
         ctx->request->nktypes = k5_count_etypes(ctx->request->ktype);
@@ -979,7 +986,7 @@ krb5_init_creds_init(krb5_context context,
                 goto cleanup;
             krb5_free_principal(context, ctx->request->client);
             ctx->request->client = new_client;
-            krb5_princ_type(context, ctx->request->client) = KRB5_NT_WELLKNOWN;
+            ctx->request->client->type = KRB5_NT_WELLKNOWN;
         }
     }
     /* We will also handle anonymous if the input principal is the anonymous
@@ -987,7 +994,7 @@ krb5_init_creds_init(krb5_context context,
     if (krb5_principal_compare_any_realm(context, ctx->request->client,
                                          krb5_anonymous_principal())) {
         ctx->request->kdc_options |= KDC_OPT_REQUEST_ANONYMOUS;
-        krb5_princ_type(context, ctx->request->client) = KRB5_NT_WELLKNOWN;
+        ctx->request->client->type = KRB5_NT_WELLKNOWN;
     }
     code = restart_init_creds_loop(context, ctx, NULL);
     if (code)
@@ -1019,7 +1026,7 @@ krb5_init_creds_set_service(krb5_context context,
     free(ctx->in_tkt_service);
     ctx->in_tkt_service = s;
 
-    krb5_preauth_request_context_fini(context);
+    k5_preauth_request_context_fini(context);
     return restart_init_creds_loop(context, ctx, NULL);
 }
 
@@ -1099,6 +1106,7 @@ init_creds_validate_reply(krb5_context context,
 static void
 read_allowed_preauth_type(krb5_context context, krb5_init_creds_context ctx)
 {
+    krb5_error_code ret;
     krb5_data config;
     char *tmp, *p;
 
@@ -1108,20 +1116,16 @@ read_allowed_preauth_type(krb5_context context, krb5_init_creds_context ctx)
     memset(&config, 0, sizeof(config));
     if (krb5_cc_get_config(context, ctx->opte->opt_private->in_ccache,
                            ctx->request->server,
-                           KRB5_CONF_PA_TYPE, &config) != 0)
+                           KRB5_CC_CONF_PA_TYPE, &config) != 0)
         return;
-    tmp = malloc(config.length + 1);
-    if (tmp == NULL) {
-        krb5_free_data_contents(context, &config);
+    tmp = k5memdup0(config.data, config.length, &ret);
+    krb5_free_data_contents(context, &config);
+    if (tmp == NULL)
         return;
-    }
-    memcpy(tmp, config.data, config.length);
-    tmp[config.length] = '\0';
     ctx->allowed_preauth_type = strtol(tmp, &p, 10);
     if (p == NULL || *p != '\0')
         ctx->allowed_preauth_type = KRB5_PADATA_NONE;
     free(tmp);
-    krb5_free_data_contents(context, &config);
 }
 
 static krb5_error_code
@@ -1138,7 +1142,7 @@ save_selected_preauth_type(krb5_context context, krb5_ccache ccache,
         return ENOMEM;
     config_data = string2data(tmp);
     code = krb5_cc_set_config(context, ccache, ctx->cred.server,
-                              KRB5_CONF_PA_TYPE, &config_data);
+                              KRB5_CC_CONF_PA_TYPE, &config_data);
     free(tmp);
     return code;
 }
@@ -1146,24 +1150,21 @@ save_selected_preauth_type(krb5_context context, krb5_ccache ccache,
 static krb5_error_code
 clear_cc_config_out_data(krb5_context context, krb5_init_creds_context ctx)
 {
-    if (ctx->cc_config_out != NULL)
-        k5_json_release(ctx->cc_config_out);
-    ctx->cc_config_out = k5_json_object_create();
-    if (ctx->cc_config_out == NULL)
-        return ENOMEM;
-    return 0;
+    k5_json_release(ctx->cc_config_out);
+    ctx->cc_config_out = NULL;
+    return k5_json_object_create(&ctx->cc_config_out);
 }
 
 static krb5_error_code
 read_cc_config_in_data(krb5_context context, krb5_init_creds_context ctx)
 {
+    k5_json_value val;
     krb5_data config;
     char *encoded;
     krb5_error_code code;
     int i;
 
-    if (ctx->cc_config_in != NULL)
-        k5_json_release(ctx->cc_config_in);
+    k5_json_release(ctx->cc_config_in);
     ctx->cc_config_in = NULL;
 
     if (ctx->opte->opt_private->in_ccache == NULL)
@@ -1172,7 +1173,7 @@ read_cc_config_in_data(krb5_context context, krb5_init_creds_context ctx)
     memset(&config, 0, sizeof(config));
     code = krb5_cc_get_config(context, ctx->opte->opt_private->in_ccache,
                               ctx->request->server,
-                              KRB5_CONF_PA_CONFIG_DATA, &config);
+                              KRB5_CC_CONF_PA_CONFIG_DATA, &config);
     if (code)
         return code;
 
@@ -1181,16 +1182,15 @@ read_cc_config_in_data(krb5_context context, krb5_init_creds_context ctx)
     if (i < 0)
         return ENOMEM;
 
-    ctx->cc_config_in = k5_json_decode(encoded);
+    code = k5_json_decode(encoded, &val);
     free(encoded);
-    if (ctx->cc_config_in == NULL)
-        return ENOMEM;
-    if (k5_json_get_tid(ctx->cc_config_in) != K5_JSON_TID_OBJECT) {
-        k5_json_release(ctx->cc_config_in);
-        ctx->cc_config_in = NULL;
+    if (code)
+        return code;
+    if (k5_json_get_tid(val) != K5_JSON_TID_OBJECT) {
+        k5_json_release(val);
         return EINVAL;
     }
-
+    ctx->cc_config_in = val;
     return 0;
 }
 
@@ -1205,12 +1205,12 @@ save_cc_config_out_data(krb5_context context, krb5_ccache ccache,
     if (ctx->cc_config_out == NULL ||
         k5_json_object_count(ctx->cc_config_out) == 0)
         return 0;
-    encoded = k5_json_encode(ctx->cc_config_out);
-    if (encoded == NULL)
-        return ENOMEM;
+    code = k5_json_encode(ctx->cc_config_out, &encoded);
+    if (code)
+        return code;
     config = string2data(encoded);
     code = krb5_cc_set_config(context, ccache, ctx->cred.server,
-                              KRB5_CONF_PA_CONFIG_DATA, &config);
+                              KRB5_CC_CONF_PA_CONFIG_DATA, &config);
     free(encoded);
     return code;
 }
@@ -1221,29 +1221,22 @@ init_creds_step_request(krb5_context context,
                         krb5_data *out)
 {
     krb5_error_code code;
-    krb5_boolean got_real;
-    char random_buf[4];
-    krb5_data random_data;
 
     if (ctx->loopcount >= MAX_IN_TKT_LOOPS) {
         code = KRB5_GET_IN_TKT_LOOP;
         goto cleanup;
     }
-    /*
-     * RFC 6113 requires a new nonce for the inner request on each try. It's
-     * permitted to change the nonce even for non-FAST so we do here.
-     */
-    random_data.length = 4;
-    random_data.data = (char *)random_buf;
-    code = krb5_c_random_make_octets(context, &random_data);
-    if (code !=0)
+
+    /* RFC 6113 requires a new nonce for the inner request on each try. */
+    code = pick_nonce(context, ctx);
+    if (code != 0)
         goto cleanup;
-    /*
-     * See RT ticket 3196 at MIT.  If we set the high bit, we may have
-     * compatibility problems with Heimdal, because we (incorrectly) encode
-     * this value as signed.
-     */
-    ctx->request->nonce = 0x7fffffff & load_32_n(random_buf);
+
+    /* Reset the request timestamps, possibly adjusting to the KDC time. */
+    code = set_request_times(context, ctx);
+    if (code != 0)
+        goto cleanup;
+
     krb5_free_data(context, ctx->inner_request_body);
     ctx->inner_request_body = NULL;
     code = encode_krb5_kdc_req_body(ctx->request, &ctx->inner_request_body);
@@ -1265,19 +1258,9 @@ init_creds_step_request(krb5_context context,
 
     if (ctx->err_reply == NULL) {
         /* either our first attempt, or retrying after PREAUTH_NEEDED */
-        code = krb5_do_preauth(context,
-                               ctx->request,
-                               ctx->inner_request_body,
-                               ctx->encoded_previous_request,
-                               ctx->preauth_to_use,
-                               &ctx->request->padata,
-                               ctx->prompter,
-                               ctx->prompter_data,
-                               &ctx->preauth_rock,
-                               ctx->opte,
-                               &got_real);
-        if (code == 0 && !got_real && ctx->preauth_required)
-            code = KRB5_PREAUTH_FAILED;
+        code = k5_preauth(context, ctx, ctx->preauth_to_use,
+                          ctx->preauth_required, &ctx->request->padata,
+                          &ctx->selected_preauth_type);
         if (code != 0)
             goto cleanup;
     } else {
@@ -1286,18 +1269,8 @@ init_creds_step_request(krb5_context context,
              * Retry after an error other than PREAUTH_NEEDED,
              * using ctx->err_padata to figure out what to change.
              */
-            code = krb5_do_preauth_tryagain(context,
-                                            ctx->request,
-                                            ctx->inner_request_body,
-                                            ctx->encoded_previous_request,
-                                            ctx->preauth_to_use,
-                                            &ctx->request->padata,
-                                            ctx->err_reply,
-                                            ctx->err_padata,
-                                            ctx->prompter,
-                                            ctx->prompter_data,
-                                            &ctx->preauth_rock,
-                                            ctx->opte);
+            code = k5_preauth_tryagain(context, ctx, ctx->preauth_to_use,
+                                       &ctx->request->padata);
         } else {
             /* No preauth supplied, so can't query the plugins. */
             code = KRB5KRB_ERR_GENERIC;
@@ -1309,6 +1282,9 @@ init_creds_step_request(krb5_context context,
         }
     }
 
+    /* Remember when we sent this request (after any preauth delay). */
+    ctx->request_time = time(NULL);
+
     if (ctx->encoded_previous_request != NULL) {
         krb5_free_data(context, ctx->encoded_previous_request);
         ctx->encoded_previous_request = NULL;
@@ -1407,7 +1383,7 @@ check_reply_enctype(krb5_init_creds_context ctx)
 /* Note the difference between the KDC's time, as reported to us in a
  * preauth-required error, and the current time. */
 static void
-note_req_timestamp(krb5_context kcontext, krb5_clpreauth_rock rock,
+note_req_timestamp(krb5_context context, krb5_init_creds_context ctx,
                    krb5_timestamp kdc_time, krb5_int32 kdc_usec)
 {
     krb5_timestamp now;
@@ -1415,29 +1391,39 @@ note_req_timestamp(krb5_context kcontext, krb5_clpreauth_rock rock,
 
     if (k5_time_with_offset(0, 0, &now, &usec) != 0)
         return;
-    rock->pa_offset = kdc_time - now;
-    rock->pa_offset_usec = kdc_usec - usec;
-    rock->pa_offset_state = (rock->fast_state->armor_key != NULL) ?
+    ctx->pa_offset = kdc_time - now;
+    ctx->pa_offset_usec = kdc_usec - usec;
+    ctx->pa_offset_state = (ctx->fast_state->armor_key != NULL) ?
         AUTH_OFFSET : UNAUTH_OFFSET;
 }
 
+/* Determine whether the client realm in a KRB-ERROR is empty. */
+static krb5_boolean
+is_empty_crealm(krb5_error *err)
+{
+
+    return (err->client == NULL || err->client->realm.length == 0);
+}
+
 /*
- * Determine whether err is a client referral to another realm, given the
- * previously requested client principal name.
+ * Determine whether a KRB-ERROR is a referral to another realm.
  *
- * RFC 6806 Section 7 requires that KDCs return the referral realm in an error
- * type WRONG_REALM, but Microsoft Windows Server 2003 (and possibly others)
- * return the realm in a PRINCIPAL_UNKNOWN message.
+ * RFC 6806 Section 7 requires that KDCs return the referral realm in
+ * an error type WRONG_REALM, but Microsoft Windows Server 2003 (and
+ * possibly others) return the realm in a PRINCIPAL_UNKNOWN message.
+ * Detect this case by looking for a non-empty client.realm field in
+ * such responses.
  */
 static krb5_boolean
-is_referral(krb5_context context, krb5_error *err, krb5_principal client)
+is_referral(krb5_init_creds_context ctx)
 {
-    if (err->error != KDC_ERR_WRONG_REALM &&
-        err->error != KDC_ERR_C_PRINCIPAL_UNKNOWN)
+    krb5_error *err = ctx->err_reply;
+
+    if (err->error == KDC_ERR_WRONG_REALM)
+        return TRUE;
+    if (err->error != KDC_ERR_C_PRINCIPAL_UNKNOWN)
         return FALSE;
-    if (err->client == NULL)
-        return FALSE;
-    return !krb5_realm_compare(context, err->client, client);
+    return !is_empty_crealm(err);
 }
 
 static krb5_error_code
@@ -1447,11 +1433,12 @@ init_creds_step_reply(krb5_context context,
 {
     krb5_error_code code;
     krb5_pa_data **kdc_padata = NULL;
+    krb5_preauthtype kdc_pa_type;
     krb5_boolean retry = FALSE;
     int canon_flag = 0;
     krb5_keyblock *strengthen_key = NULL;
     krb5_keyblock encrypting_key;
-    krb5_boolean fast_avail, got_real;
+    krb5_boolean fast_avail;
 
     encrypting_key.length = 0;
     encrypting_key.contents = NULL;
@@ -1473,7 +1460,7 @@ init_creds_step_reply(krb5_context context,
             goto cleanup;
         if (negotiation_requests_restart(context, ctx, ctx->err_padata)) {
             ctx->have_restarted = 1;
-            krb5_preauth_request_context_fini(context);
+            k5_preauth_request_context_fini(context);
             if ((ctx->fast_state->fast_state_flags & KRB5INT_FAST_DO_FAST) ==0)
                 ctx->enc_pa_rep_permitted = 0;
             code = restart_init_creds_loop(context, ctx, ctx->err_padata);
@@ -1487,9 +1474,9 @@ init_creds_step_reply(krb5_context context,
             krb5_free_pa_data(context, ctx->preauth_to_use);
             ctx->preauth_to_use = ctx->err_padata;
             ctx->err_padata = NULL;
-            note_req_timestamp(context, &ctx->preauth_rock,
-                               ctx->err_reply->stime, ctx->err_reply->susec);
-            /* this will trigger a new call to krb5_do_preauth() */
+            note_req_timestamp(context, ctx, ctx->err_reply->stime,
+                               ctx->err_reply->susec);
+            /* This will trigger a new call to k5_preauth(). */
             krb5_free_error(context, ctx->err_reply);
             ctx->err_reply = NULL;
             code = sort_krb5_padata_sequence(context,
@@ -1497,18 +1484,22 @@ init_creds_step_reply(krb5_context context,
                                              ctx->preauth_to_use);
             ctx->preauth_required = TRUE;
 
-        } else if (canon_flag && is_referral(context, ctx->err_reply,
-                                             ctx->request->client)) {
+        } else if (canon_flag && is_referral(ctx)) {
+            if (is_empty_crealm(ctx->err_reply)) {
+                /* Only WRONG_REALM referral types can reach this. */
+                code = KRB5KDC_ERR_WRONG_REALM;
+                goto cleanup;
+            }
             TRACE_INIT_CREDS_REFERRAL(context, &ctx->err_reply->client->realm);
             /* Rewrite request.client with realm from error reply */
             krb5_free_data_contents(context, &ctx->request->client->realm);
             code = krb5int_copy_data_contents(context,
                                               &ctx->err_reply->client->realm,
                                               &ctx->request->client->realm);
-            /* this will trigger a new call to krb5_do_preauth() */
+            /* This will trigger a new call to k5_preauth(). */
             krb5_free_error(context, ctx->err_reply);
             ctx->err_reply = NULL;
-            krb5_preauth_request_context_fini(context);
+            k5_preauth_request_context_fini(context);
             /* Permit another negotiation based restart. */
             ctx->have_restarted = 0;
             ctx->sent_nontrivial_preauth = 0;
@@ -1538,7 +1529,7 @@ init_creds_step_reply(krb5_context context,
         goto cleanup;
 
     /* process any preauth data in the as_reply */
-    krb5_clear_preauth_context_use_counts(context);
+    k5_reset_preauth_types_tried(context);
     code = krb5int_fast_process_response(context, ctx->fast_state,
                                          ctx->reply, &strengthen_key);
     if (code != 0)
@@ -1551,26 +1542,11 @@ init_creds_step_reply(krb5_context context,
 
     ctx->etype = ctx->reply->enc_part.enctype;
 
-    /*
-     * At this point, allow whichever preauth plugin that can handle the KDC's
-     * reply padata to do so, regardless of that data's padata type.  We don't
-     * want to record the type of padata in the reply, so set the pointer for
-     * that data to NULL.
-     */
+    /* Process the final reply padata.  Don't restrict the preauth types or
+     * record a selected preauth type. */
     ctx->allowed_preauth_type = KRB5_PADATA_NONE;
-    ctx->preauth_rock.selected_preauth_type = NULL;
-
-    code = krb5_do_preauth(context,
-                           ctx->request,
-                           ctx->inner_request_body,
-                           ctx->encoded_previous_request,
-                           ctx->reply->padata,
-                           &kdc_padata,
-                           ctx->prompter,
-                           ctx->prompter_data,
-                           &ctx->preauth_rock,
-                           ctx->opte,
-                           &got_real);
+    code = k5_preauth(context, ctx, ctx->reply->padata, FALSE, &kdc_padata,
+                      &kdc_pa_type);
     if (code != 0)
         goto cleanup;
 
@@ -1653,8 +1629,7 @@ init_creds_step_reply(krb5_context context,
     if (code)
         goto cleanup;
 
-    code = stash_as_reply(context, ctx->request_time, ctx->request,
-                          ctx->reply, &ctx->cred, NULL);
+    code = stash_as_reply(context, ctx->reply, &ctx->cred, NULL);
     if (code != 0)
         goto cleanup;
     if (ctx->opte && ctx->opte->opt_private->out_ccache) {
@@ -1670,7 +1645,7 @@ init_creds_step_reply(krb5_context context,
             config_data.data = "yes";
             config_data.length = strlen(config_data.data);
             code = krb5_cc_set_config(context, out_ccache, ctx->cred.server,
-                                      KRB5_CONF_FAST_AVAIL, &config_data);
+                                      KRB5_CC_CONF_FAST_AVAIL, &config_data);
             if (code != 0)
                 goto cc_cleanup;
         }
@@ -1688,9 +1663,10 @@ init_creds_step_reply(krb5_context context,
         }
     }
 
-    krb5_preauth_request_context_fini(context);
+    k5_preauth_request_context_fini(context);
 
     /* success */
+    code = 0;
     ctx->complete = TRUE;
 
 cleanup:
@@ -1781,18 +1757,12 @@ cleanup:
 }
 
 krb5_error_code KRB5_CALLCONV
-krb5int_get_init_creds(krb5_context context,
-                       krb5_creds *creds,
-                       krb5_principal client,
-                       krb5_prompter_fct prompter,
-                       void *prompter_data,
-                       krb5_deltat start_time,
-                       const char *in_tkt_service,
-                       krb5_get_init_creds_opt *options,
-                       krb5_gic_get_as_key_fct gak_fct,
-                       void *gak_data,
-                       int  *use_master,
-                       krb5_kdc_rep **as_reply)
+k5_get_init_creds(krb5_context context, krb5_creds *creds,
+                  krb5_principal client, krb5_prompter_fct prompter,
+                  void *prompter_data, krb5_deltat start_time,
+                  const char *in_tkt_service, krb5_get_init_creds_opt *options,
+                  get_as_key_fn gak_fct, void *gak_data, int *use_master,
+                  krb5_kdc_rep **as_reply)
 {
     krb5_error_code code;
     krb5_init_creds_context ctx = NULL;
@@ -1836,10 +1806,10 @@ cleanup:
 }
 
 krb5_error_code
-krb5int_populate_gic_opt(krb5_context context, krb5_get_init_creds_opt **out,
-                         krb5_flags options, krb5_address *const *addrs,
-                         krb5_enctype *ktypes,
-                         krb5_preauthtype *pre_auth_types, krb5_creds *creds)
+k5_populate_gic_opt(krb5_context context, krb5_get_init_creds_opt **out,
+                    krb5_flags options, krb5_address *const *addrs,
+                    krb5_enctype *ktypes, krb5_preauthtype *pre_auth_types,
+                    krb5_creds *creds)
 {
     int i;
     krb5_int32 starttime;
